/**
 * This app plays voice prompts during a voice call as follows:
 *
 * - Play VOICE_PROMPT_START voice prompt (just once);
 * - Play DIALING audio file (in loop);
 * - Initiate the voice call;
 * - As soon as the call event is ALERTING, stop DIALING audio file playback and play RINGTONE audio
 *   file (in loop);
 * - As soon as the call event is CONNECTED, stop RINGTONE audio file playback and speak/listen to
 *   the phone;
 * - Hangup the call;
 * - As soon as the call event is TERMINATED, play VOICE_PROMPT_END voice prompt (just once).
 *
 * Audio interfaces are the analog ones (microphone and speaker).
 *
 * You must issue the following commands:
 * @verbatim
   $ app start voicePromptMcc
   $ app runProc voicePromptMcc --exe=voicePromptMcc -- <phone number>
   $ app runProc voicePromptMcc --exe=voicePromptMcc -- <phone number> AMR-NB for playing AMR-NB encoded voice
   prompts
   @endverbatim
 *
 * @note If Ctrl-C is issued while the call is connected, the last voice prompt (VOICE_PROMPT_END)
 *       won't be played since we exit before the playback starts.
 *
 * Copyright (C) Sierra Wireless Inc. Use of this work is subject to license.
 *
 */

#include "legato.h"
#include <pthread.h>
#include <unistd.h>
#include "interfaces.h"
#include <opencore-amrnb/interf_dec.h>

//--------------------------------------------------------------------------------------------------
// Symbol and Enum definitions.
//--------------------------------------------------------------------------------------------------
#define VOICE_PROMPT_START  "/usr/share/sounds/SwiECallStartMono"
#define VOICE_PROMPT_END    "/usr/share/sounds/SwiECallCanceledMono"
#define DIALING             "/usr/share/sounds/SwiDialingMono"
#define RINGTONE            "/usr/share/sounds/SwiRingBackToneFrMono"


#define BUFFER_SIZE         2048
#define AMR_DECODER_BUFFER_LEN  500

//--------------------------------------------------------------------------------------------------
// Data structures.
//--------------------------------------------------------------------------------------------------

typedef struct {
    uint32_t riffId;         ///< "RIFF" constant. Marks the file as a riff file.
    uint32_t riffSize;       ///< Size of the overall file - 8 bytes
    uint32_t riffFmt;        ///< File Type Header. For our purposes, it always equals "WAVE".
    uint32_t fmtId;          ///< Equals "fmt ". Format chunk marker. Includes trailing null
    uint32_t fmtSize;        ///< Length of format data as listed above
    uint16_t audioFormat;    ///< Audio format (PCM)
    uint16_t channelsCount;  ///< Number of channels
    uint32_t sampleRate;     ///< Sample frequency in Hertz
    uint32_t byteRate;       ///< sampleRate * channelsCount * bps / 8
    uint16_t blockAlign;     ///< channelsCount * bps / 8
    uint16_t bitsPerSample;  ///< Bits per sample
    uint32_t dataId;         ///< "data" chunk header. Marks the beginning of the data section.
    uint32_t dataSize;       ///< Data size
} WavHeader_t;


typedef enum
{
    MODE_WAV,
    MODE_AMR_NB,
    MODE_MAX
}
DecodingMode_t;



//--------------------------------------------------------------------------------------------------
//                                       Static declarations
//--------------------------------------------------------------------------------------------------

static le_mcc_CallRef_t TestCallRef;

static le_audio_StreamRef_t MdmRxAudioRef = NULL;
static le_audio_StreamRef_t MdmTxAudioRef = NULL;
static le_audio_StreamRef_t FeInRef = NULL;
static le_audio_StreamRef_t FeOutRef = NULL;
static le_audio_StreamRef_t PlayerRef = NULL;

static le_audio_ConnectorRef_t AudioInputConnectorRef = NULL;
static le_audio_ConnectorRef_t AudioOutputConnectorRef = NULL;

static le_audio_MediaHandlerRef_t MediaHandlerRef = NULL;

static int   AudioFileFd = -1;
static int   Pipefd[2];

static le_thread_Ref_t PlaySamplesRef = NULL;

static void* AmrHandle = NULL;

static void* PlaySamplesThread (void* contextPtr);
static void* AmrDecoderThread (void* contextPtr);

static struct
{
    char*                   extension;
    le_thread_MainFunc_t    threadFunc;
}
DecodingModeData[MODE_MAX]=
{
    {
        "wav",
        PlaySamplesThread
    },
    {
        "amr",
        AmrDecoderThread
    }
};


static DecodingMode_t DecodingMode = MODE_WAV;

/* From WmfDecBytesPerFrame in dec_input_format_tab.cpp */
static const int AmrNbSizes[] = { 12, 13, 15, 17, 19, 20, 26, 31, 5, 6, 5, 5, 0, 0, 0, 0 };

//--------------------------------------------------------------------------------------------------
/**
 * Disconnect all streams and connectors
 *
 */
//--------------------------------------------------------------------------------------------------
static void DisconnectAllAudio
(
    void
)
{
    if (AudioInputConnectorRef)
    {
        if(PlayerRef)
        {
            LE_INFO("Disconnect %p from connector.%p", PlayerRef, AudioInputConnectorRef);
            le_audio_Disconnect(AudioInputConnectorRef, PlayerRef);
        }
        if (FeInRef)
        {
            LE_INFO("Disconnect %p from connector.%p", FeInRef, AudioInputConnectorRef);
            le_audio_Disconnect(AudioInputConnectorRef, FeInRef);
        }
        if(MdmTxAudioRef)
        {
            LE_INFO("Disconnect %p from connector.%p", MdmTxAudioRef, AudioInputConnectorRef);
            le_audio_Disconnect(AudioInputConnectorRef, MdmTxAudioRef);
        }
    }
    if(AudioOutputConnectorRef)
    {
        if(PlayerRef)
        {
            LE_INFO("Disconnect %p from connector.%p", PlayerRef, AudioOutputConnectorRef);
            le_audio_Disconnect(AudioOutputConnectorRef, PlayerRef);
        }
        if(FeOutRef)
        {
            LE_INFO("Disconnect %p from connector.%p", FeOutRef, AudioOutputConnectorRef);
            le_audio_Disconnect(AudioOutputConnectorRef, FeOutRef);
        }
        if(MdmRxAudioRef)
        {
            LE_INFO("Disconnect %p from connector.%p", MdmRxAudioRef, AudioOutputConnectorRef);
            le_audio_Disconnect(AudioOutputConnectorRef, MdmRxAudioRef);
        }
    }

    if(AudioInputConnectorRef)
    {
        le_audio_DeleteConnector(AudioInputConnectorRef);
        AudioInputConnectorRef = NULL;
    }
    if(AudioOutputConnectorRef)
    {
        le_audio_DeleteConnector(AudioOutputConnectorRef);
        AudioOutputConnectorRef = NULL;
    }

    if(PlayerRef)
    {
        le_audio_Close(PlayerRef);
        PlayerRef = NULL;
    }
    if(FeInRef)
    {
        le_audio_Close(FeInRef);
        FeInRef = NULL;
    }
    if(FeOutRef)
    {
        le_audio_Close(FeOutRef);
        FeOutRef = NULL;
    }
    if(MdmRxAudioRef)
    {
        le_audio_Close(MdmRxAudioRef);
        MdmRxAudioRef = NULL;
    }
    if(MdmTxAudioRef)
    {
        le_audio_Close(MdmTxAudioRef);
        MdmTxAudioRef = NULL;
    }

    close(AudioFileFd);
    close(Pipefd[0]);
    close(Pipefd[1]);

    if ( DecodingMode == MODE_AMR_NB )
    {
        Decoder_Interface_exit(AmrHandle);
    }
    exit(EXIT_SUCCESS);
}

//--------------------------------------------------------------------------------------------------
/**
 *  Stop playback.
 *
 */
//--------------------------------------------------------------------------------------------------
static void StopFilePlayback
(
    void
)
{
    if (PlaySamplesRef)
    {
        le_thread_Cancel(PlaySamplesRef);
        PlaySamplesRef = NULL;
    }

    if(PlayerRef)
    {
        le_audio_Flush(PlayerRef);
    }

    close(AudioFileFd);
}

//--------------------------------------------------------------------------------------------------
/**
 *  Play Samples thread destructor.
 *
 */
//--------------------------------------------------------------------------------------------------
static void DestroyPlayThread
(
    void *contextPtr
)
{
    PlaySamplesRef = NULL;
}


//--------------------------------------------------------------------------------------------------
/**
 *  AMR decoder thread.
 *
 */
//--------------------------------------------------------------------------------------------------
static void* AmrDecoderThread
(
    void* contextPtr
)
{
    LE_DEBUG("AMR decoding start");

    lseek(AudioFileFd, 0, SEEK_SET);

    char header[10] = {0};
    uint32_t totalRead=6;

    if ( read(AudioFileFd, header, 6) != 6 )
    {
        LE_ERROR("AMR detection: cannot read header");
        return NULL;
    }

    if ( strncmp(header, "#!AMR\n", 6) != 0 )
    {
        LE_ERROR("Unsupported format %s", header);
        return NULL;
    }

    while (1)
    {
        uint8_t outBuffer[320];
        int16_t tmpBuff[160];
        uint8_t readBuff[AMR_DECODER_BUFFER_LEN];
        int i,n,size;


        /* Read the mode byte */
        n = read(AudioFileFd, readBuff, 1);
        if (n <= 0)
        {
            LE_DEBUG("No more data to read");
            break;
        }

        /* Find the packet size */
        size = AmrNbSizes[(readBuff[0] >> 3) & 0x0f];

        if ((size < 0) || (size > (AMR_DECODER_BUFFER_LEN-1)))
        {
            break;
        }

        totalRead+=1+size;

        n = read(AudioFileFd, readBuff + 1, size);
        if (n != size)
        {
            LE_ERROR("Underflow in AMR decoding");
            break;
        }

        /* Decode the packet */
        Decoder_Interface_Decode( AmrHandle, readBuff, tmpBuff, 0);

        /* Convert to little endian and write to wav */
        for (i = 0; i < 160; i++)
        {
            outBuffer[2*i] = (tmpBuff[i] >> 0) & 0xff;
            outBuffer[2*i+1] = (tmpBuff[i] >> 8) & 0xff;
        }


        write(Pipefd[1], outBuffer, 320);
    }

    LE_DEBUG("AMR decoding end %d", totalRead);

    return NULL;
}

//--------------------------------------------------------------------------------------------------
/**
 *  Play Samples thread.
 *
 */
//--------------------------------------------------------------------------------------------------
static void* PlaySamplesThread
(
    void* contextPtr
)
{
    char data[BUFFER_SIZE];
    int32_t len = 1;

    lseek(AudioFileFd, 0+sizeof(WavHeader_t), SEEK_SET);

    LE_INFO("Read audio file...");

    while ((len = read(AudioFileFd, data, BUFFER_SIZE)) > 0)
    {
        int32_t wroteLen = write(Pipefd[1], data, len);
        if (wroteLen <= 0)
        {
            LE_ERROR("write error errno.%d (%s)", errno, strerror(errno));
            return NULL;
        }
    }

    LE_INFO("End of audio file reached");

    return NULL;
}

//--------------------------------------------------------------------------------------------------
/**
 *  Play Samples.
 *
 */
//--------------------------------------------------------------------------------------------------
static void PlaySamples
(
    void
)
{
    if (!PlaySamplesRef)
    {
        PlaySamplesRef = le_thread_Create( "PlaySamples",
                                            DecodingModeData[DecodingMode].threadFunc,
                                            NULL);

        le_thread_AddChildDestructor(PlaySamplesRef,
                                     DestroyPlayThread,
                                     NULL);

        le_thread_Start(PlaySamplesRef);
    }
}

//--------------------------------------------------------------------------------------------------
/**
 * Play a file
 *
 */
//--------------------------------------------------------------------------------------------------
static void PlayFile
(
    const char* audioFilePathPtr
)
{
    char filePathPtr[100]={0};
    snprintf(filePathPtr, 100, "%s.%s", audioFilePathPtr, DecodingModeData[DecodingMode].extension);

    if ((AudioFileFd=open(filePathPtr, O_RDONLY)) == -1)
    {
        LE_ERROR("Open file %s failure: errno.%d (%s)",  filePathPtr, errno, strerror(errno));
        DisconnectAllAudio();
        exit(EXIT_FAILURE);
    }
    else
    {
        LE_INFO("Open file %s with AudioFileFd.%d",  filePathPtr, AudioFileFd);
    }

    PlaySamples();
}

//--------------------------------------------------------------------------------------------------
/**
 * Play file automaton
 *
 */
//--------------------------------------------------------------------------------------------------
static void PlayFileOnEvent
(
    le_mcc_Event_t           callEvent,
    le_audio_MediaEvent_t    mediaEvent
)
{
    typedef enum
    {
        STAT_NONE,
        STATE_START,
        STATE_DIALING,
        STATE_RINGTONE,
        STATE_PROMPT_END,
        STATE_MAX
    }
    PlayState_t;

    static struct
    {
        const char* path;
        bool playInLoop;
    }
    playInfo[STATE_MAX] =
    {
        { NULL,               false },
        { VOICE_PROMPT_START, false },
        { DIALING,            true  },
        { RINGTONE,           true  },
        { VOICE_PROMPT_END,   false }
    };

    static PlayState_t currentState = STAT_NONE;

    LE_INFO( "callEvent %d mediaEvent %d currentState %d", callEvent, mediaEvent, currentState );

    if (mediaEvent == LE_AUDIO_MEDIA_NO_MORE_SAMPLES)
    {
        LE_INFO("playInLoop %d", playInfo[currentState].playInLoop);

        if ( playInfo[currentState].playInLoop )
        {
            PlaySamples();
        }
        else
        {
            if ( currentState == STATE_START )
            {
                le_mcc_Start(TestCallRef);
            }

            currentState++;

            LE_DEBUG("currentState %d", currentState);

            if ( currentState < STATE_PROMPT_END )
            {
                PlayFile(playInfo[currentState].path);
            }
            else if (currentState == STATE_MAX)
            {
                // Test ended
                LE_INFO("Test ends successfully.");
                exit(EXIT_SUCCESS);
            }
        }
    }

    switch (callEvent)
    {
        case LE_MCC_EVENT_ALERTING:
        case LE_MCC_EVENT_CONNECTED:
            // Stop to play in loop
            playInfo[currentState].playInLoop = false;
            StopFilePlayback();
        break;

        case LE_MCC_EVENT_TERMINATED:
            StopFilePlayback();
            currentState = STATE_PROMPT_END;
            PlayFile(playInfo[currentState].path);
        break;
        default:
        break;
    }
}

//--------------------------------------------------------------------------------------------------
/**
 * Handler function for Stream Event Notifications.
 *
 */
//--------------------------------------------------------------------------------------------------
static void MyMediaEventHandler
(
    le_audio_StreamRef_t          streamRef,
    le_audio_MediaEvent_t         event,
    void*                         contextPtr
)
{
    switch(event)
    {
        case LE_AUDIO_MEDIA_NO_MORE_SAMPLES:
            LE_INFO("Media event is LE_AUDIO_MEDIA_NO_MORE_SAMPLES.");
        break;

        case LE_AUDIO_MEDIA_ENDED:
            LE_INFO("Media event is LE_AUDIO_MEDIA_ENDED.");
        break;

        case LE_AUDIO_MEDIA_ERROR:
            LE_INFO("Media event is LE_AUDIO_MEDIA_ERROR.");
        break;
        default:
            LE_INFO("Media event is %d", event);
        break;
    }

    PlayFileOnEvent(LE_MCC_EVENT_MAX, event);
}

//--------------------------------------------------------------------------------------------------
/**
 * Connect audio
 *
 */
//--------------------------------------------------------------------------------------------------
static void ConnectAudio
(
    void
)
{
    // Redirect audio to the in-built Microphone and Speaker.
    FeOutRef = le_audio_OpenSpeaker();
    LE_ERROR_IF((FeOutRef==NULL), "OpenSpeaker returns NULL!");
    FeInRef = le_audio_OpenMic();
    LE_ERROR_IF((FeInRef==NULL), "OpenMic returns NULL!");

    AudioInputConnectorRef = le_audio_CreateConnector();
    LE_ERROR_IF((AudioInputConnectorRef==NULL), "AudioInputConnectorRef is NULL!");
    AudioOutputConnectorRef = le_audio_CreateConnector();
    LE_ERROR_IF((AudioOutputConnectorRef==NULL), "AudioOutputConnectorRef is NULL!");

    LE_ERROR_IF((le_audio_Connect(AudioInputConnectorRef, FeInRef)!=LE_OK),
                "Failed to connect Mic on Input connector!");
    LE_ERROR_IF((le_audio_Connect(AudioOutputConnectorRef, FeOutRef)!=LE_OK),
                "Failed to connect Speaker on Output connector!");

    MdmRxAudioRef = le_audio_OpenModemVoiceRx();
    LE_ERROR_IF((MdmRxAudioRef==NULL), "GetRxAudioStream returns NULL!");
    MdmTxAudioRef = le_audio_OpenModemVoiceTx();
    LE_ERROR_IF((MdmTxAudioRef==NULL), "GetTxAudioStream returns NULL!");

    // Play local on output connector.
    PlayerRef = le_audio_OpenPlayer();
    LE_ERROR_IF((PlayerRef==NULL), "OpenFilePlayback returns NULL!");

    // Set profile and specific gains for AR7/AR8 (won't work on other platforms)
    LE_ERROR_IF((le_audio_SetProfile(1) != LE_OK), "Cannot set profile 1");
    LE_ERROR_IF((le_audio_SetGain(PlayerRef, 0x300) != LE_OK), "Cannot set multimedia gain");
    LE_ERROR_IF((le_audio_SetGain(MdmRxAudioRef, 5) != LE_OK), "Cannot set MdmRxAudioRef gain");

    MediaHandlerRef = le_audio_AddMediaHandler(PlayerRef, MyMediaEventHandler, NULL);
    LE_ERROR_IF((MediaHandlerRef==NULL), "AddMediaHandler returns NULL!");

    LE_ERROR_IF((le_audio_Connect(AudioOutputConnectorRef, PlayerRef) != LE_OK),
                "Failed to connect FilePlayback on output connector!");

    LE_ERROR_IF((pipe(Pipefd) == -1), "Failed to create the pipe");

    if (DecodingMode == MODE_WAV)
    {
        LE_ERROR_IF((le_audio_SetSamplePcmChannelNumber(PlayerRef, 1) != LE_OK),
                    "Cannot set the channel number");
        LE_ERROR_IF((le_audio_SetSamplePcmSamplingRate(PlayerRef, 8000) != LE_OK),
                    "Cannot set the sampling rate");
        LE_ERROR_IF((le_audio_SetSamplePcmSamplingResolution(PlayerRef, 16) != LE_OK),
                    "Cannot set the sampling resolution");
    }
    else if (DecodingMode == MODE_AMR_NB)
    {
        LE_ERROR_IF((le_audio_SetSamplePcmChannelNumber(PlayerRef, 1) != LE_OK),
                    "Cannot set the channel number");
        LE_ERROR_IF((le_audio_SetSamplePcmSamplingRate(PlayerRef, 8000) != LE_OK),
                    "Cannot set the sampling rate");
        LE_ERROR_IF((le_audio_SetSamplePcmSamplingResolution(PlayerRef, 16) != LE_OK),
                    "Cannot set the sampling resolution");
    }

    LE_ERROR_IF((le_audio_PlaySamples(PlayerRef, Pipefd[0]) != LE_OK), "Cannot play samples");
}

//--------------------------------------------------------------------------------------------------
/**
 * Handler function for Call Event Notifications.
 *
 */
//--------------------------------------------------------------------------------------------------
static void MyCallEventHandler
(
    le_mcc_CallRef_t  callRef,
    le_mcc_Event_t    callEvent,
    void*             contextPtr
)
{
    le_result_t         res;

    PlayFileOnEvent(callEvent, LE_AUDIO_MEDIA_MAX);

    switch (callEvent)
    {
        case LE_MCC_EVENT_ALERTING:
        {
            LE_INFO("Call event is LE_MCC_EVENT_ALERTING.");
            break;
        }
        case LE_MCC_EVENT_CONNECTED:
        {
            LE_INFO("Call event is LE_MCC_EVENT_CONNECTED.");

            // Connect voice call to audio
            res = le_audio_Connect(AudioInputConnectorRef, MdmTxAudioRef);
            LE_ERROR_IF((res!=LE_OK), "Failed to connect mdmTx on Input connector!");
            res = le_audio_Connect(AudioOutputConnectorRef, MdmRxAudioRef);
            LE_ERROR_IF((res!=LE_OK), "Failed to connect mdmRx on Output connector!");

            break;
        }
        case LE_MCC_EVENT_TERMINATED:
        {
            LE_INFO("Call event is LE_MCC_EVENT_TERMINATED.");
            le_mcc_TerminationReason_t term = le_mcc_GetTerminationReason(callRef);
            switch(term)
            {
                case LE_MCC_TERM_LOCAL_ENDED:
                case LE_MCC_TERM_REMOTE_ENDED:
                    LE_INFO("Termination reason is LE_MCC_TERM_REMOTE_ENDED or LE_MCC_TERM_LOCAL_ENDED");
                    break;

                case LE_MCC_TERM_NETWORK_FAIL:
                case LE_MCC_TERM_UNASSIGNED_NUMBER:
                case LE_MCC_TERM_USER_BUSY:
                case LE_MCC_TERM_UNDEFINED:
                default:
                    LE_INFO("Termination reason is %d", term);
            }

            le_mcc_Delete(callRef);

            break;
        }
        case LE_MCC_EVENT_INCOMING:
        {
            LE_INFO("Call event is LE_MCC_EVENT_INCOMING.");
            break;
        }
        default:
        {
            LE_INFO("Unknowm Call event.");
            break;
        }
    }
}

//--------------------------------------------------------------------------------------------------
/**
 * Helper.
 *
 */
//--------------------------------------------------------------------------------------------------
static void PrintUsage()
{
    int idx;
    bool sandboxed = (getuid() != 0);
    const char * usagePtr[] = {
            "Usage of the voicePromptMcc test is:",
            "   \"app runProc voicePromptMcc --exe=voicePromptMcc -- <phone number>\" with wav file",
            "   \"app runProc voicePromptMcc --exe=voicePromptMcc -- <phone number> AMR-NB\" with AMR-NB \
decoding file"
    };

    for(idx = 0; idx < NUM_ARRAY_MEMBERS(usagePtr); idx++)
    {
        if(sandboxed)
        {
            LE_INFO("%s", usagePtr[idx]);
        }
        else
        {
            fprintf(stderr, "%s\n", usagePtr[idx]);
        }
    }
}

//--------------------------------------------------------------------------------------------------
/**
 * The signal event handler function for SIGINT/SIGTERM when process dies.
 */
//--------------------------------------------------------------------------------------------------
static void SigHandler
(
    int sigNum
)
{
    if (TestCallRef)
    {
        LE_INFO("HangUp call");
        le_mcc_HangUp(TestCallRef);
    }
    // If Ctrl-C is issued while the call is connected, the last voice prompt (VOICE_PROMPT_END)
    // won't be played since we exit before the playback starts.
    exit(EXIT_SUCCESS);
}

//--------------------------------------------------------------------------------------------------
/**
 * Initialize the test component.
 * Start application with 'app start voicePromptMcc' command
 * Execute application with 'app runProc voicePromptMcc --exe=voicePromptMcc (see PrintUsage())'
 */
//--------------------------------------------------------------------------------------------------
COMPONENT_INIT
{
    const char* destinationNumber;

    if ((le_arg_NumArgs() >= 1) && (le_arg_NumArgs() <= 2))
    {
        // Register a signal event handler for SIGINT when user interrupts/terminates process
        signal(SIGINT, SigHandler);

        LE_INFO("======== Start voicePromptMcc Test ========");

        destinationNumber = le_arg_GetArg(0);

        if ( le_arg_NumArgs() == 2 )
        {
            const char* mode = le_arg_GetArg(1);

            if ( strcmp(mode, "AMR-NB") == 0 )
            {
                DecodingMode = MODE_AMR_NB;
                AmrHandle = Decoder_Interface_init();
            }
        }

        // Connect audio stuff
        ConnectAudio();

        // Prepare call handling
        le_mcc_AddCallEventHandler(MyCallEventHandler, NULL);
        TestCallRef = le_mcc_Create(destinationNumber);

        // Start playing voice prompt once
        PlayFileOnEvent(LE_MCC_EVENT_MAX, LE_AUDIO_MEDIA_NO_MORE_SAMPLES);

        LE_INFO("======== voicePromptMcc Test started successfully ========");
    }
    else
    {
        PrintUsage();
        LE_INFO("EXIT voicePromptMcc Test");
        exit(EXIT_FAILURE);
    }
}

