//--------------------------------------------------------------------------------------------------
/**
 * @page mkTools_NinjaScriptGenerator Ninja Script Generator
 *
 * There are various ninja build scripts that get generated by this module.
 *
 * There's a function to generate each type of .ninja:
 *
 * - ninja::Generate(const Component_t*) - Stand-alone component library.
 *
 * - ninja::Generate(const Exe_t*) - Executable and associated component libraries.
 *
 * - ninja::Generate(const App_t*) - Application bundle, including executables, component libraries,
 *                                   bundled files, a configuration data file,
 *                                   and an info.properties file.
 *
 * - ninja::Generate(const System_t*) - System bundle, including all app bundles and user binding
 *                                      configuration.
 *
 * None of the scripts do configuration generation or exe main code generation.  The scripts expect
 * those things to be done before the scripts are run.  The ninja scripts will run ifgen to
 * generate interface code, though.
 *
 * Each script includes a rule to check that none of the definition files that it is generated
 * from have been changed.  If they have, the script will delete itself and run the appropriate
 * mktool to regenerate itself.
 *
 * <hr>
 *
 * Copyright (C) Sierra Wireless Inc.  Use of this work is subject to license.
 */
//--------------------------------------------------------------------------------------------------

#ifndef LEGATO_NINJA_SCRIPT_GENERATOR_H_INCLUDE_GUARD
#define LEGATO_NINJA_SCRIPT_GENERATOR_H_INCLUDE_GUARD

namespace ninja
{


#include "toolChain.h"


//--------------------------------------------------------------------------------------------------
/**
 * Generate a build script for a stand-alone component library.
 **/
//--------------------------------------------------------------------------------------------------
void Generate
(
    const model::Component_t* componentPtr,
    const mk::BuildParams_t& buildParams,
    int argc,           ///< Count of the number of command line parameters.
    const char** argv   ///< Pointer to array of pointers to command line argument strings.
);


//--------------------------------------------------------------------------------------------------
/**
 * Generate a build script for an executable and associated component and IPC interface
 * libraries.
 **/
//--------------------------------------------------------------------------------------------------
void Generate
(
    const model::Exe_t* exePtr,
    const mk::BuildParams_t& buildParams,
    int argc,           ///< Count of the number of command line parameters.
    const char** argv   ///< Pointer to array of pointers to command line argument strings.
);


//--------------------------------------------------------------------------------------------------
/**
 * Generate a build script for an application.
 **/
//--------------------------------------------------------------------------------------------------
void Generate
(
    const model::App_t* appPtr,
    const mk::BuildParams_t& buildParams,
    const std::string& outputDir,   ///< Path to the directory into which the built app will be put.
    int argc,           ///< Count of the number of command line parameters.
    const char** argv   ///< Pointer to array of pointers to command line argument strings.
);


//--------------------------------------------------------------------------------------------------
/**
 * Generate a build script for a module.
 **/
//--------------------------------------------------------------------------------------------------
void Generate
(
    const model::Module_t* modulePtr,
    const mk::BuildParams_t& buildParams,
    int argc,           ///< Count of the number of command line parameters.
    const char** argv   ///< Pointer to array of pointers to command line argument strings.
);


//--------------------------------------------------------------------------------------------------
/**
 * Generate a build script for a system.
 **/
//--------------------------------------------------------------------------------------------------
void Generate
(
    const model::System_t* systemPtr,
    const mk::BuildParams_t& buildParams,
    const std::string& outputDir,   ///< Path to the directory where the built system will be put.
    int argc,           ///< Count of the number of command line parameters.
    const char** argv   ///< Pointer to an array of pointers to command line argument strings.
);


} // namespace ninja

#endif // LEGATO_NINJA_SCRIPT_GENERATOR_H_INCLUDE_GUARD
